@@ -84,11 +84,12 @@ typedef unsigned long t_ul;
 /**  Data storage for ignore patterns. */
 struct ignore_t {
 	/** The pattern string as given by the user, including flags. */
-	char *pattern,
-			 /** The calculated pattern string.
-				* Does no longer include the flags (like \e take), and shell syntax 
-				* is converted to PCRE. */
-			 *compare_string;
+	char *pattern;
+
+	/** The calculated pattern string.
+	 * Does no longer include the flags (like \e take), and shell syntax 
+	 * is converted to PCRE. */
+	char *compare_string;
 
 	union {
 		/* for shell and pcre */
@@ -111,7 +112,7 @@ struct ignore_t {
 			unsigned short path_level;
 			/** Flag telling whether this shell pattern has a \c ** in it.
 			 * \todo Would be used with ignore_t::path_level. */
-			int has_wildwildcard:1;
+			unsigned int has_wildwildcard:1;
 		};
 
 		/** For device compares */
@@ -136,17 +137,17 @@ struct ignore_t {
 		};
 	};
 
-	/** Is this an ignore or take pattern? 
-	 * \a 0 = take, \a 1 = ignore */
-	int is_ignore:1;
+	/** Is this an ignore or take pattern?  \a 0 = take, \a 1 = ignore */
+	unsigned int is_ignore:1;
 	/** Ignore case for comparing? */
-	int is_icase:1;
+	unsigned int is_icase:1;
 	/** Is it an \e internally generated pattern (for the WAA area)?
 	 * Internal patterns are not saved and not printed. */
-	int is_user_pat:1;
-	/** Which type is this pattern? \see Pattern types. */
+	unsigned int is_user_pat:1;
+
+	/** Which type is this pattern? See \ref PatTypes. */
 	/* This is at the end because of alignment issues. */
-	unsigned char type;
+	unsigned int type:3;
 };
 
 /** Whether the device compare should be 
@@ -172,17 +173,22 @@ struct ignore_t {
 struct url_t {
 	/** The URL itself (http:// or svn:// or similar) */
 	char *url;
-	/** The length of the URL, not counting the \c \\0. */
-	int urllen;
-	/** The revision we'd like that URL to be at - normally HEAD. */
- 	svn_revnum_t target_rev;
-	/** The revision number this URL is currently at. */
-	svn_revnum_t current_rev;
-	/** The user-given symbolic name */
-	char *name;
 	/** The user-given priority; need not be unique. 
 	 * The lower the number, the higher the priority. */
 	int priority; 
+	/** The length of the URL, not counting the \c \\0. */
+	int urllen;
+	/** The revision we'd like that URL to be at - normally HEAD. */
+	svn_revnum_t target_rev;
+	/** The revision the user gave <B>for this command</b> for this URL.
+	 * Normally equals \c target_rev. */
+	svn_revnum_t current_target_rev;
+	/** The revision number this URL is currently at. */
+	svn_revnum_t current_rev;
+	/** The \c HEAD revision, or \c SVN_INVALID_REVNUM if not yet known. */
+	svn_revnum_t head_rev;
+	/** The user-given symbolic name */
+	char *name;
 	/** The number which is used in the dir-lists to reference this url_t.
 	 * Must be unique in the URL-list. 
 	 *
@@ -200,7 +206,9 @@ struct url_t {
 	apr_pool_t *pool;
 	/** Flag saying whether this URL should be done.
 	 * Should not be queried directly, but by using url__to_be_handled().  */
-	int to_be_handled;
+	int to_be_handled:1;
+	/** Whether the user gave a specific override revision number. */
+	int current_target_override:1;
 };
 
 
@@ -221,27 +229,6 @@ struct url_t {
  * no alignment problem (which would arise if there was eg. a 16bit type,
  * then space would be wasted) */
 struct sstat_t {
-	/** Device number */
-	dev_t dev;
-	/** Inode */
-	ino_t ino;
-
-	union {
-		/** The size in bytes (for files, symlinks and directories). */
-		off_t	size;
-		/** The device number (for devices). */
-		dev_t rdev;
-	};
-
-	/** The access mode (like \c 0700, \c 0755) with all other (non-mode) 
-	 * bits, ie S_IFDIR. */
-	mode_t mode;
-
-	/** The owner's id. */
-	uid_t uid;
-	/** The group number. */
-	gid_t gid;
-
 	/* For easier comparison, we overlay an 64bit type. */
 	union {
 		/** The modification time as \c seconds, \c microseconds. */
@@ -257,6 +244,27 @@ struct sstat_t {
 		 * \deprecated Currently unused. */
 		unsigned long long _ctime;
 	};
+
+	union {
+		/** The size in bytes (for files, symlinks and directories). */
+		off_t	size;
+		/** The device number (for devices). */
+		dev_t rdev;
+	};
+
+	/** Device number of \b host filesystem. */
+	dev_t dev;
+	/** Inode */
+	ino_t ino;
+
+	/** The access mode (like \c 0700, \c 0755) with all other (non-mode) 
+	 * bits, ie S_IFDIR. */
+	mode_t mode;
+
+	/** The owner's id. */
+	uid_t uid;
+	/** The group number. */
+	gid_t gid;
 };
 
 
@@ -265,53 +273,53 @@ struct sstat_t {
  *
  * The name comes from <em>extended struct stat</em>.
  * This structure is used to build the tree of entries that we're processing.
- *
- * \note Some moving of members may be necessary to get optimal alignment. */
+ */
 struct estat {
-	/** Meta-data of this entry. */
-	struct sstat_t st;
-
+	/** The parent of this entry, used for tree walking.
+	 * Must be \c NULL for the root entry and the root entry alone. */
+	struct estat *parent;
 	/** Name of this entry. */
 	char *name;
 
+	/** Meta-data of this entry. */
+	struct sstat_t st;
+
 	/** Revision of this entry. Currently only the value in the root entry is 
 	 * used; this will be moved to \c * \ref url and removed from here. */
 	svn_revnum_t repos_rev;
 	/** The revision number before updating. */
 	svn_revnum_t old_rev;
 
-	/** The parent of this entry, used for tree walking.
-	 * Must be \c NULL for the root entry and the root entry alone. */
-	struct estat *parent;
-
 	/** The URL this entry is from.
 	 * Will be used for multi-url updates. */
 	struct url_t *url;
 
+	/** If an entry gets removed, the old version is remembered (if needed) 
+	 * via the \c old pointer (eg to know which children were known and may 
+	 * be safely removed).  */
+	struct estat *old;
+
 	/** Data about this entry. */
-	struct {
+	union {
 		/** For files */
 		struct {
 			/** MD5-hash of the repository version. 
 			 * While committing it is set to the \e new MD5, and saved 
 			 * with \a waa__output_tree(). */
 			md5_digest_t md5;
-			/** Flag whether this entry has changed or not changed 
-			 * (as per MD5/manber-compare), or if this is unknown yet.
-			 * See \ref ChgFlag. */
-			int change_flag;
 			/** The decoder string from fsvs:update-pipe. Only gets set if 
 			 * action->needs_decoder != 0. */
 			char *decoder;
 			/** Whether we got an "original" MD5 from the repository to compare.  
 			 * */
-			int has_orig_md5:1;
+			unsigned int has_orig_md5:1;
+			/** Flag whether this entry has changed or not changed (as per 
+			 * MD5/manber-compare), or if this is unknown yet.
+			 * See \ref ChgFlag. */
+			unsigned int change_flag:2;
 		};
 		/** For directories */
 		struct {
-			/** How many entries this directory has. */
-			unsigned entry_count;
-
 			/** List of child entries.
 			 * Sorted by inode number, NULL-terminated.
 			 * Always valid. */
@@ -325,13 +333,17 @@ struct estat {
 			 * Mainly used in the root inode, but is used in newly found directories
 			 * too. \c NULL for most directory entries. */
 			char *strings;
+
+			/** How many entries this directory has. */
+			unsigned entry_count;
+
 			/** This flag is set if any child is *not* at the same revision,
 			 * so this directory has to be descended on reporting. */
-			int other_revs:1;
+			unsigned int other_revs:1;
 			/** If this bit is set, the directory has to be re-sorted before
 			 * being written out -- it may have new entries, which are not in
 			 * the correct order. */
-			int to_be_sorted:1;
+			unsigned int to_be_sorted:1;
 			/* Currently unused - see ignore.c. */
 #if 0
 			struct ignore_t **active_ign;
@@ -347,28 +359,21 @@ struct estat {
 			/** This entries' baton. */
 			void *baton;
 		};
-		/** Update for a file. */
+		/** Export for a file. */
 		struct {
 			/** The pool used for the filehandles; for a discussion see \ref FHP. */
 			apr_pool_t *filehandle_pool;
 		};
-		/** Update for a special entry. */
+		/** Export of a special entry. */
 		struct {
 			/** String-buffers for special entries.
 			 * While a file is \b always streamed to disk, special entries are
 			 * \b always done in memory. */
-			svn_stringbuf_t *stringbuf_src, 
-											*stringbuf_tgt;
+			svn_stringbuf_t *stringbuf_tgt;
 		};
 		struct {
 			/** Used in waa__input_tree() and waa__update_tree(). */
 			unsigned child_index;
-			/** Local pool while updating a directory.
-			 * For the children in a directory we have to create subpools in \ref 
-			 * up__apply_textdelta(); the local \c pool argument is too 
-			 * short-lived, they have to have at least the lifetime of their 
-			 * parent directories. */
-			apr_pool_t *dir_pool;
 		};
 		/* output_tree */
 		struct {
@@ -378,13 +383,19 @@ struct estat {
 
 
 	/** \name Common variables for all types of entries. */
+	/** Which argument causes this path to be done. */
+	char *arg;
+
+	/** Stored user-defined properties as \c name=>svn_string_t, if \c 
+	 * action->keep_user_prop is set.
+	 * Allocated in a subpool of \c estat::url->pool, so that it's still 
+	 * available after cb__record_changes() returns.
+	 * The subpool is available from a hash lookup with key "" (len=0). */
+  apr_hash_t *user_prop;
+
 	/** Flags for this entry. See \ref EntFlags for constant definitions. */
 	int flags;
 
-	/** Length of path up to here. Does not include the \c \\0. See \ref 
-	 * ops__calc_path_len. */
-	unsigned short path_len:14;
-
 	/** What kind of entry this is - see \ref ent_types. */
 	unsigned int entry_type:6;
 	/** Local status of this entry - \ref fs_bits. */
@@ -398,30 +409,32 @@ struct estat {
 	 * value here has range of <tt>[1 .. MAX_CACHED_PATHS]</tt> instead of 
 	 * the usual <tt>[0 .. MAX_CACHED_PATHS-1]</tt>. */
 	unsigned int cache_index:6;
+
+	/** Length of path up to here. Does not include the \c \\0. See \ref 
+	 * ops__calc_path_len. */
+	unsigned short path_len:16;
+
 	/** At which level is this path? The wc root is at level 0, its children 
 	 * at 1, and so on. */
 	unsigned short path_level:9;
 
 	/** Whether this entry was already printed. \todo Remove by changing the 
 	 * logic. */
-	int was_output:1;
+	unsigned int was_output:1;
 	/** This flag tells whether the string for the decoder is surely correct.
 	 * It is currently used for updates; after we parse the properties in 
 	 * cb__record_changes(), we'll have the correct value. */
-	int decoder_is_correct:1;
+	unsigned int decoder_is_correct:1;
 
 	/** Flag saying whether this entry was specified by the user on the 
 	 * command line. */
-	int do_tree:1;
+	unsigned int do_tree:1;
 	/** Like \a FS_CHILD_CHANGED - children of this entry must be handled.
 	 * \todo substitute by \a FS_CHILD_CHANGED? */
-	int do_a_child:1;
+	unsigned int do_a_child:1;
 	/** Flag derived from parents' \ref estat::do_tree.
 	 * Set for \b all entries which should be handled. */
-	int do_this_entry:1;
-
-	/** Which argument causes this path to be done. */
-	char *arg;
+	unsigned int do_this_entry:1;
 };
 
 
@@ -563,42 +576,42 @@ struct estat {
 /** Declaration of the debug function. */
 extern void _DEBUGP(const char *file, int line, const char *func, char *format, ...)
 	__attribute__ ((format (printf, 4, 5) ));
-/** The macro used for printing debug messages.
- * Includes time, file, line number and function name. 
- * Allows filtering via opt_debugprefix.
- * \note Check for \ref PrintfTypes "argument sizes". */
+	/** The macro used for printing debug messages.
+	 * Includes time, file, line number and function name. 
+	 * Allows filtering via opt_debugprefix.
+	 * \note Check for \ref PrintfTypes "argument sizes". */
 #define DEBUGP(...) _DEBUGP(__FILE__, __LINE__, __PRETTY_FUNCTION__, __VA_ARGS__)
 #endif
 
 
-/** \name Error-printing and -handling functions. 
- *
- * Except for the subversion-library wrapper macros they need exactly this
- * function layout:
- *
- * \code
- *   int some_function( ... some args ... )
- *   {
- *     int status;
- *
- *     STOPIF( function_call(1, 2, "a"), 
- *       "String describing the error situation with %s",
- *       "parameters as needed");
- *
- *   ex:
- *     cleanups();
- *     return status;
- *   }
- * \endcode
- *
- * It works by checking the return value; if it is not zero, a 
- * <tt>goto ex</tt> is done. At this mark some cleanup is possible. */
-/** @{ */
-/** A flag to turn error printing temporarily off.
- * This is useful where entire calltrees would have to be equipped with 
- * some \c silent parameter. */
-extern int make_STOP_silent;
-/** Master error function. */
+	/** \name Error-printing and -handling functions. 
+	 *
+	 * Except for the subversion-library wrapper macros they need exactly this
+	 * function layout:
+	 *
+	 * \code
+	 *   int some_function( ... some args ... )
+	 *   {
+	 *     int status;
+	 *
+	 *     STOPIF( function_call(1, 2, "a"), 
+	 *       "String describing the error situation with %s",
+	 *       "parameters as needed");
+	 *
+	 *   ex:
+	 *     cleanups();
+	 *     return status;
+	 *   }
+	 * \endcode
+	 *
+	 * It works by checking the return value; if it is not zero, a 
+	 * <tt>goto ex</tt> is done. At this mark some cleanup is possible. */
+	/** @{ */
+	/** A flag to turn error printing temporarily off.
+	 * This is useful where entire calltrees would have to be equipped with 
+	 * some \c silent parameter. */
+	extern int make_STOP_silent;
+	/** Master error function. */
 extern int _STOP(const char *file, int line, const char *function,
 		int errl, const char *format, ...)
 __attribute__ ((format (printf, 5, 6) ));
@@ -637,14 +650,14 @@ __attribute__ ((format (printf, 5, 6) ));
  * \param code The status code to check.
  * All other things are hardcoded. */
 #define STOPIF(code, ... ) \
-	do                                                  \
+	do                                                \
 {                                                   \
-	status=code;                                      \
-	if (status)                                       \
-	{                                                 \
+	status=(code);                                    \
+	if (status)                     									\
+	{																									\
 		_STOP(__FILE__, __LINE__, __PRETTY_FUNCTION__, status, __VA_ARGS__);   \
 		goto ex;                                        \
-	}                                                 \
+	}																									\
 } while (0)
 /** A simplified error call macro for returning ENOMEM.
  * \code
@@ -655,6 +668,21 @@ __attribute__ ((format (printf, 5, 6) ));
  * \endcode
  * */
 #define STOPIF_ENOMEM(cond) STOPIF_CODE_ERR(cond, ENOMEM, NULL)
+/** An error return macro that is used for user output - special handling 
+ * \c EPIPE to get a silent return.
+ * If \c code returns something negative (like printf, puts, putc ... do; 
+ * \c EOF is defined as \c -1), and \a error is \c EPIPE, go on with \c 
+ * -EPIPE. */
+#define STOPIF_CODE_EPIPE(code, ...) \
+	do                                                   \
+{                                                      \
+	if ((code) < 0)                                      \
+	{                                                    \
+		status=errno;                                      \
+		if (status == EPIPE) status= -EPIPE;               \
+		STOPIF(status, "Error writing output");          \
+	}                                                    \
+} while (0)
 
 /** \page svnlibwrap Subversion library calls wrapper.
  * If this is used in some function, an additional variable is needed:
@@ -676,18 +704,12 @@ __attribute__ ((format (printf, 5, 6) ));
  * \endcode
  */ 
 /** The master error macro for calling subversion functions. */
-#define STOPIF_SVNERR_EXTRA(func, parm, fmt, ...)\
-	do                                                  \
-{                                                   \
-	status_svn=func parm;                             \
-	if (status_svn)                                   \
-	{                                                 \
-		status=status_svn->apr_err;											\
-		_STOP(__FILE__, __LINE__, __PRETTY_FUNCTION__, 	\
-				status_svn->apr_err, 												\
-				fmt ": %s", ## __VA_ARGS__, status_svn->message);\
-		goto ex;                                        \
-	}                                                 \
+#define STOPIF_SVNERR_EXTRA(func, parm, fmt, ...)          \
+	do                                                       \
+{                                                          \
+	status_svn=func parm;                                    \
+	STOPIF_CODE_ERR( status_svn, status_svn->apr_err,   \
+			fmt ": %s", ## __VA_ARGS__, status_svn->message);  \
 } while (0)
 /* The mainly used function wrapper.
  * \param func Name of the subversion function
@@ -696,7 +718,7 @@ __attribute__ ((format (printf, 5, 6) ));
  *   STOPIF_SVNERR( svn_ra_initialize, (global_pool));
  * \endcode
  */
-#define STOPIF_SVNERR(func, parm) STOPIF_SVNERR_TEXT(func, #func, parm)
+#define STOPIF_SVNERR(func, parm) STOPIF_SVNERR_EXTRA(func, parm, #func)
 /** The same as STOPIF_SVNERR(), but with a variable printed function name.
  * Used in case the function is called indirectly; see STOPIF_SVNERR_INDIR. */
 #define STOPIF_SVNERR_TEXT(func, text, parm) STOPIF_SVNERR_EXTRA(func, parm, text)
@@ -763,93 +785,93 @@ __attribute__ ((format (printf, 5, 6) ));
  * (and the modification time check is not enough).
  * Makes checking slower, as data gets MD5ed. */
 extern int opt_checksum, 
-	/** Greater than zero if additional details are wanted, 
-	* or negative for extra quiet operation. */
-	opt_verbose, 
-	/** Flag for recursive/non-recursive behaviour.
-	 * Starting with 0, gets incremented with \c -R and decremented with \c 
-	 * -N. Different actions have different default levels. */
-	opt_recursive,
-	/** If this is an import/export command (eg restoration after harddisk 
-	 * crash), we don't use the WAA for data storage. */
-	is_import_export,
-	/** Flag saying whether the local update should only set the entry_status
-	* of existing entries and not check for new ones. Needed for update. */
-	only_check_status,
-	/** Whether debug messages are wanted. */
-	debuglevel;
+			 /** Greater than zero if additional details are wanted, 
+				* or negative for extra quiet operation. */
+			 opt_verbose, 
+			 /** Flag for recursive/non-recursive behaviour.
+				* Starting with 0, gets incremented with \c -R and decremented with \c 
+				* -N. Different actions have different default levels. */
+			 opt_recursive,
+			 /** If this is an import/export command (eg restoration after harddisk 
+				* crash), we don't use the WAA for data storage. */
+			 is_import_export,
+			 /** Flag saying whether the local update should only set the entry_status
+				* of existing entries and not check for new ones. Needed for update. */
+			 only_check_status,
+			 /** Whether debug messages are wanted. */
+			 debuglevel;
 
-/** A pointer to the commit message; possibly a mmap()ped file. */
-extern char *opt_commitmsg,
-	/** The file name of the commit message file. */
-	*opt_commitmsgfile;
+			 /** A pointer to the commit message; possibly a mmap()ped file. */
+			 extern char *opt_commitmsg,
+			 /** The file name of the commit message file. */
+			 *opt_commitmsgfile;
 
-/** The revision we're getting from the repository. */
-extern svn_revnum_t target_revision;
-/** The revision the user wants to get at (\c -r parameter).
-* \c HEAD is represented by \c SVN_INVALID_REVNUM.
-* Has to be splitted per-URL when we're going to multi-url operation. */
-extern svn_revnum_t opt_target_revision;
-/** The second revision number the user specified. */
-extern svn_revnum_t opt_target_revision2;
-/** How many revisions the user specified on the commandline (0, 1 or 2).
-* For multi-update operations it's possibly to update the urls to different
-* revisions; then we need to know for which urls the user specified a
-* revision number. Per default we go to \c HEAD.
-* */
-extern int opt_target_revisions_given;
+			 /** The revision we're getting from the repository. */
+			 extern svn_revnum_t target_revision;
+			 /** The revision the user wants to get at (\c -r parameter).
+				* \c HEAD is represented by \c SVN_INVALID_REVNUM.
+				* Has to be splitted per-URL when we're going to multi-url operation. */
+			 extern svn_revnum_t opt_target_revision;
+			 /** The second revision number the user specified. */
+			 extern svn_revnum_t opt_target_revision2;
+			 /** How many revisions the user specified on the commandline (0, 1 or 2).
+				* For multi-update operations it's possibly to update the urls to different
+				* revisions; then we need to know for which urls the user specified a
+				* revision number. Per default we go to \c HEAD.
+				* */
+			 extern int opt_target_revisions_given;
 
-/** The local character encoding, according to \a LC_ALL or \a LC_CTYPE) */
+			 /** The local character encoding, according to \a LC_ALL or \a LC_CTYPE) */
 #ifdef HAVE_LOCALES
-extern char *local_codeset;
+			 extern char *local_codeset;
 #endif
 
-/** The session handle for RA operations. */
-extern svn_ra_session_t *session;
+			 /** The session handle for RA operations. */
+			 extern svn_ra_session_t *session;
 
-/** The first allocated APR pool. All others are derived from it and its
-* children. */
-extern apr_pool_t *global_pool;
+			 /** The first allocated APR pool. All others are derived from it and its
+				* children. */
+			 extern apr_pool_t *global_pool;
 
-/** The array of URLs. */
-extern struct url_t **urllist;
-/** Number of URLs we have. */
-extern int urllist_count;
-/** Pointer to \b current URL. */
-extern struct url_t *current_url;
+			 /** The array of URLs. */
+			 extern struct url_t **urllist;
+			 /** Number of URLs we have. */
+			 extern int urllist_count;
+			 /** Pointer to \b current URL. */
+			 extern struct url_t *current_url;
 
-extern unsigned approx_entry_count;
-/** @} */
+			 extern unsigned approx_entry_count;
+			 /** @} */
 
 
-extern char propname_mtime[],
-	/** Modification time - \c svn:owner */
-	propname_owner[],
-	/** Modification time - \c svn:group */
-	propname_group[],
-	/** Modification time - \c svn:unix-mode */
-	propname_umode[],
-	/** Original MD5 for encoded entries. */
-	propname_origmd5[],
-	/** Flag for special entry. */
-	propname_special[],
-	/** The value for the special property; normally \c "*". */
-	propval_special[],
+			 extern char propname_mtime[],
+			 /** Modification time - \c svn:owner */
+			 propname_owner[],
+			 /** Modification time - \c svn:group */
+			 propname_group[],
+			 /** Modification time - \c svn:unix-mode */
+			 propname_umode[],
+			 /** Original MD5 for encoded entries. */
+			 propname_origmd5[],
+			 /** Flag for special entry. */
+			 propname_special[],
+			 /** The value for the special property; normally \c "*". */
+			 propval_special[],
 
-	/** Commit-pipe program.  */
-	propval_commitpipe[],
-	/** Update-pipe program. */
-	propval_updatepipe[];
+			 /** Commit-pipe program.  */
+			 propval_commitpipe[],
+			 /** Update-pipe program. */
+			 propval_updatepipe[];
 
 
-/** \addtogroup cmds_strings Common command line strings
- * \ingroup compat
- *
- * These strings may have to be localized some time, that's why they're
- * defined in this place. */
-/** @{ */
-extern char parm_dump[],
-		 parm_load[];
+			 /** \addtogroup cmds_strings Common command line strings
+				* \ingroup compat
+				*
+				* These strings may have to be localized some time, that's why they're
+				* defined in this place. */
+			 /** @{ */
+			 extern char parm_dump[],
+			 parm_load[];
 /** @} */
 
 /** Remember where we started. */
